##############################################################################################################
#									
# Routine :	alm_syncprojects.pl	v1.1			
#   									
# Description :								
#	This perl script copies Phase, Role, and WorkConfiguration records from a source ALMProject
#   to a target ALMProject.
#									
# Syntax :								
#	ratlperl alm_syncprojects.pl {-projadmin <projadmin>} {-pw <pw>} {-dbname <db>} {-dbset <dbset>} {-fromprojid <projectid>} {-toprojid <projectid>}
#
#	example: ratlperl alm_syncprojects.pl -projadmin ProjectAdmin -pw "" -dbname CQALM -dbset TestSchema -fromprojid CQALM00000001 -toprojid CQALM00000002
#											
#
# Licensed Materials - Property of IBM
# Rational ClearQuest
# (C) Copyright IBM Corp. 2009.  All Rights Reserved
# US Government Users Restricted Rights
# Use, duplication or disclosure restricted by GSA ADP Schedule Contract # with IBM Corp
###############################################################################################################									
#
# Change Log:
#
# Ver       CRs             Date         Programmer   Comments (fixed, fixed with dependency, added feature)
#
# 1.1	    RATLC01477119   04/27/2009   majidi	      RFE to support ALMProject sync
#
# Known issues:
##############################################################################################################
#!ratlperl -l -w
use CQPerlExt;
use strict;
use File::Basename;
use Getopt::Long;

##################
# Global variables
my $progname = basename $0;
my $entity;
my %APHash=();
use subs qw(diffProjects cpEngine cpRecords CQValidateCommit Usage);

#############################
# Define the input parameters

my @options = qw(
    projadmin=s		
    pw=s		
    dbname=s		
    dbset=s	
    fromprojid=s	
    toprojid=s
    help);

my %options = (
    projadmin => undef,
    pw => undef,
    dbname => undef,
    dbset => undef,
    fromprojid => undef,
    toprojid => undef,
);
######################
# Get the input parameters

unless (GetOptions(\%options, @options)) {
    my $retval = &Usage;
    exit 0;
}
########################
# Display the help usage

if ($options{help}) {
	print "-projadmin       ProjectAdmin ĵ¼ʶ\n";
	print "-pw              ProjectAdmin 롣\n";
	print "-dbname          ûݿơ\n";
	print "-dbset           ݿݿ⼯ӵơ\n";
	print "-fromprojid      ƲԴ ALMProject ʶ\n";
	print "-toprojid        ƲĿ ALMProject ʶ\n";
	print "-help            ڰĿѡ־\n";
	exit 0;
}

######################################################
# End program if any of the parameters are not defined

     if(!defined $options{projadmin}
     or !defined $options{pw}
     or !defined $options{dbname}
     or !defined $options{dbset}
     or !defined $options{fromprojid}
     or !defined $options{toprojid})
  { 
     #Printing Usage to the command line
     my $retval = &Usage();
     exit 1;
  }
#################################
# Subroutine to display the usage
sub Usage {
	 print "\nͬ ALMProject ʵó\n
	
	$progname
	{-projadmin <ĿԱ>}
	{-pw <>}
	{-dbname <ݿ>}
	{-dbset <ݿ⼯>}
	{-fromprojid <ԴĿʶ>}	
	{-toprojid <ĿĿʶ>}
	[-help]
";
return 0;
}

###############
# main

my $FromProjId=$options{fromprojid};
my $ToProjId=$options{toprojid};
my $session = CQSession::Build();

#####################
# Establish a session
eval{$session->UserLogon($options{projadmin},$options{pw},$options{dbname},$options{dbset});};
die "** ¼ʱ $options{projadmin},$options{pw},$options{dbname},$options{dbset} \n$@\n" if ($@);

###############
# Validate the Ids provided for projects
eval{$session->GetEntity("ALMProject","$FromProjId");};
if ($@){
	CQPerlExt::CQSession_Unbuild($session);
	die "** Ŀʱ $FromProjId  GetEntity ʧ" if ($@);
}
eval{$session->GetEntity("ALMProject","$ToProjId");};
if ($@){
	CQPerlExt::CQSession_Unbuild($session);
	die "** Ŀʱ $ToProjId  GetEntity ʧ" if ($@);
}
cpEngine(diffProjects("ALMPhase"), "ALMPhase");
restorePriorPhase();
cpEngine(diffProjects("ALMRole"), "ALMRole");
cpEngine(diffProjects("ALMWorkConfiguration"), "ALMWorkConfiguration");

#####################
# Unbuild the session
CQPerlExt::CQSession_Unbuild($session);
exit 0;

###############################################################################################
# sub diffProjects - Compares two ALMProjects for a given record type and returns a diff array

sub diffProjects {
	
	my($recordType) = @_;
	my %ToProjRecs;
	my @FromProjRecs;
	my @Diff;	
	my $operator;
	my $resultset;
	my $count;
	my $size;
	my $uniqueKey;
	my $querydef;
	my $buildField;
	my $ALMProject="Project";
	my @newDiff;
		
	if ("$recordType" eq "ALMRole"){$buildField="RoleLabel";}
	elsif ("$recordType" eq "ALMWorkConfiguration"){$buildField="ALMType";$ALMProject="ALMProject";}
	elsif ("$recordType" eq "ALMPhase"){$buildField="PhaseLabel";}
	
	# Query the Copy To Project
	##################################################
	
	$querydef = $session->BuildQuery("$recordType");
	$querydef->BuildField("$buildField");
	$operator = $querydef->BuildFilterOperator($CQPerlExt::CQ_BOOL_OP_AND);
	$operator->BuildFilter("$ALMProject", $CQPerlExt::CQ_COMP_OP_EQ, ["$ToProjId"]);
	$operator->BuildFilter("Obsolete", $CQPerlExt::CQ_COMP_OP_NEQ,["Yes"]); 
	$resultset = $session->BuildResultSet($querydef);
	$resultset->EnableRecordCount();			
	$resultset->Execute;	
	$count = $resultset->GetRecordCount();			
	while ($resultset->MoveNext eq $CQPerlExt::CQ_SUCCESS) {		
		$ToProjRecs{$resultset->GetColumnValue(1)} = 1;		
	}
	
	# Query the Copy from Project
	##################################################
	
	$querydef->BuildField("$buildField");
	$operator = $querydef->BuildFilterOperator($CQPerlExt::CQ_BOOL_OP_AND);
	$operator->BuildFilter("$ALMProject", $CQPerlExt::CQ_COMP_OP_EQ, ["$FromProjId"]);
	$operator->BuildFilter("Obsolete", $CQPerlExt::CQ_COMP_OP_NEQ,["Yes"]); 
	$resultset = $session->BuildResultSet($querydef);
	$resultset->EnableRecordCount();			
	$resultset->Execute;	
	$count = $resultset->GetRecordCount();
	if ($count == 0){			
		print("\nδ $FromProjId ҵκҪƵļ¼\n");		
	}
	else {
		while ($resultset->MoveNext eq $CQPerlExt::CQ_SUCCESS) {		
			push @FromProjRecs,$resultset->GetColumnValue(1);		
		}
	}	
	foreach (@FromProjRecs){
		push(@Diff,$_) unless exists $ToProjRecs{$_};
	}
	if (($#Diff > -1) && ($recordType eq "ALMWorkConfiguration")){
		push(@newDiff,grep(/ALMActivity/, @Diff));
		push(@newDiff,grep(/ALMTask/, @Diff)); 
		push(@newDiff,grep(/ALMRequest/, @Diff));
		@Diff=@newDiff;		
	}
	if ($#Diff == -1){
		print("\nδ $FromProjId ҵҪƵ $ToProjId κµ $recordType\n\n");	
	}
	return (\@Diff);
}
###########################################################################
# sub restorePriorPhase - Restore the values for ALMPhase record on pass II

sub restorePriorPhase {
	
	my $refh_AP=\%APHash;
	my $APEntity;
	my $failure;
	my $priorPhaseLabel;
	
	print("ڵڶָ׶Ρ¼ֵ...\n") if (keys %APHash);
	foreach my $AP ( keys %$refh_AP ) {    
       $priorPhaseLabel="";
       $priorPhaseLabel =  $refh_AP->{$AP}->{'PriorPhaseLabel'};
       print("ALMPhase Ϊ$AP\n");
       if ($priorPhaseLabel) {
	 		print("PriorPhaseLabel Ϊ $priorPhaseLabel\n");
	 		$APEntity = $session->GetEntity("ALMPhase","$AP");
	 		$APEntity->EditEntity("modify");
	 		$failure=$APEntity->SetFieldValue("PriorPhaseLabel","$priorPhaseLabel");
       }
       CQValidateCommit($APEntity);
	}
}
#######################################################################################
# sub cpEngine - Call to cpRecords for every element of the array

sub cpEngine {
	my($ref_diffArray,$recType ) = @_;
	my @Diff=@$ref_diffArray;
	my $uniqueKey;
	
	if ($#Diff > -1) {
	 print(" $recType Ŀ$FromProjIdƵĿ$ToProjId\n");
	 foreach (@Diff){
		$uniqueKey="$FromProjId $_";
		print("$uniqueKey\n");
		cpRecords("$uniqueKey", "$recType");
	 }
	}
}
#######################################################################################
# sub CQValidateCommit - Validates or commits a record for phase II updated
sub CQValidateCommit {
	
	my $Entity=shift(@_);
	my $status;
	print("֤ʵ\n");
	$status="";
	my $currEntityName = $Entity->GetDisplayName();
	my $currEntityType = $Entity->GetEntityDefName();
	eval {$status = $Entity->Validate()};
	if  ($@ || ($status ne "")) 
	{
		my $msg = "\n֤ʧ\n$@\nص״̬Ϊ$status\n$currEntityType$currEntityName\nѻԭ\n";
		print("$msg");
	    $Entity->Revert();
	    CQPerlExt::CQSession_Unbuild($session);
		die "$msg";
	}
	else
	{
		print("ʵʵ\n");
		eval {$status = $Entity->Commit()};
		if ($@ || ($status ne ""))
		{
			my $msg = "\nʵʧ\n$@\nص״̬Ϊ$status\n$currEntityType$currEntityName\nѻԭ\n";
			print("$msg");
	    	$Entity->Revert();
	    	CQPerlExt::CQSession_Unbuild($session);
			die "$msg";
	
		}
		else {
			my $msg = "ѳɹʵ $currEntityType$currEntityName\n";
			print("$msg");
		}

	}
	return $status;
}
#######################################################################################
# sub cprecords - Performs the record copy

sub cpRecords {	
    my($result);
    my($uniqueKey,$recordType) = @_;
	
	# This function implements ClearQuest's CopyEntity API
	# Please note that the aforementioned API is called internally only.
	# It is neither published nor supported, and this implementation should not be imitated.

	my @copyArray=();
	my @skipArray=();
	my $idx;
	my $entsCount;
	my $status;
	my $strarray1;
	my $strarray2;
	my $ent;
	my $dspname;
	my $primary;
	my $secondary;
	my $roles;
	my $primaryRole;
	my @ptypes=();
	my @stypes=();
	my @AWCroles=();
	my $currentPhaseLabel;
	my $phaseLabel;
	my $localreplica;
	my $members;
	my $AREntity;
	my @ARMembers=();
	my $origdspName;
	my $phase;
	my $priorPhaseLabel;
	my %opts=();
	my $failure;
	my $entityobjs;
	my $actions;
	my @ARActions=();
	my @ARGroups=();
	my $groups;
	my ($querydef,$operator,$resultset,$count);
	my $Entity;
	my @ARMembers;


	
	# Need a Entity Object for project
	eval {$Entity = $session->GetEntity("$recordType","$uniqueKey");};
	if ($@){
		CQPerlExt::CQSession_Unbuild($session);
		die "\n** ¼ʱ $uniqueKey  GetEntity ʧ\n$@\n";
	}	
	$localreplica= ($session->GetLocalReplica())->GetDisplayName(); 

	####################
	# Call to CopyEntity
	print("ڵ CopyEntity $@\n");

	@copyArray=('$recordType');

	##Skip Array

	@skipArray = ('ALMPhas.NextPhase'
	,'ALMPhase.CurrentIteration'
	,'ALMPhase.CurrentIterationLabel'
	,'ALMPhase.StatusLabel');


	eval {$entityobjs = $session->CopyEntity($Entity, \@copyArray, \@skipArray);};
	if ($@) {
	 CQPerlExt::CQSession_Unbuild($session);
	 die "** ʹ CopyEntity Ƽ¼ʱ\n$@";
	}

	$entsCount = $entityobjs->Count();
	print(" CopyEntity дʵ = $entsCount\n");

	########################################################################################
	#If you are copying a stateless record make sure the keys are unique and will not cause 
	#duplicates.
	for ($idx=0; $idx < $entsCount; $idx++) {

		$ent = $entityobjs->Item ($idx);
		my $entdefname = $ent->GetEntityDefName();
		my $dspName = $ent->GetDisplayName();
		print ("ʵ DefName Ϊ$entdefname \n DisplayName Ϊ$dspName\n");

		if($entdefname eq "ALMPhase")
		{	
			$dspName =~ s/$FromProjId/$ToProjId/;
			$priorPhaseLabel=$ent->GetFieldStringValue("PriorPhaseLabel");
			if ($priorPhaseLabel) {
			   	$APHash{$dspName}{'PriorPhaseLabel'}=$priorPhaseLabel;
			}
	
			$failure=$ent->SetFieldValue("PriorPhaseLabel","");
			$failure=$ent->SetFieldValue("PriorPhase","");
			$failure=$ent->SetFieldValue("Description","");
			$failure=$ent->SetFieldValue("StartDate","");
			$failure=$ent->SetFieldValue("EndDate","");
			$failure=$ent->SetFieldValue("ratl_mastership","$localreplica");
			$failure=$ent->SetFieldValue("Project","$ToProjId");
	
		}
		if($entdefname eq "ALMRole")
		{
			$members=$actions=$groups=undef;
			$origdspName = $dspName;
			$origdspName =~ s/$ToProjId/$FromProjId/;
			$AREntity = $session->GetEntity("ALMRole","$origdspName");
			$members  =	$AREntity->GetFieldStringValue("Members");
			$groups   =	$AREntity->GetFieldStringValue("Groups");
			$actions  = $AREntity->GetFieldStringValue("ApprovedActions");

			$failure=$ent->SetFieldValue("Description","");
			$failure=$ent->SetFieldValue("EntityName","");
			$failure=$ent->SetFieldValue("Members","");
			$failure=$ent->SetFieldValue("ApprovedActions","");
			$failure=$ent->SetFieldValue("ratl_mastership","$localreplica");
			$failure=$ent->SetFieldValue("Project","$ToProjId");
	
       		if ($members || $actions || $groups) {
	 			print("Members ֵ\n");
         		if ($members) {
					@ARMembers = split(/\n/, $members);
	 				foreach (@ARMembers) {
						print("Ա$_\n");
						$failure=$ent->AddFieldValue("Members","$_");
	 				} 
         		}
				if ($groups) {
	 				@ARGroups = split(/\n/, $groups);
					foreach (@ARGroups) {
						print("$_\n");
						$failure=$ent->AddFieldValue("Groups","$_");
	 				} 
        		}
	 			if ($actions) { 
         			@ARActions = split(/ /, $actions);
	 				foreach (@ARActions) {
						print("׼Ĳ\n$_\n");
						$failure=$ent->AddFieldValue("ApprovedActions","$_");
	 				}
        		}
       		}
		}
		if($entdefname eq "ALMWorkConfiguration")
		{	
			$primary = $secondary = $roles = $primaryRole = undef;
			$primary = $ent->GetFieldStringValue("ChildrenTypes_primary");
			$secondary = $ent->GetFieldStringValue("ChildrenTypes_secondary");
			$roles = $ent->GetFieldStringValue("ALMRoles");
			$primaryRole = $ent->GetFieldStringValue("PrimaryRole");
			
			$failure=$ent->SetFieldValue("ChildrenTypes_primary","");
			$failure=$ent->SetFieldValue("ChildrenTypes_secondary","");
			$failure=$ent->SetFieldValue("ALMRoles","");
			$failure=$ent->SetFieldValue("PrimaryRole","");
			$failure=$ent->SetFieldValue("ratl_mastership","$localreplica");
			$failure=$ent->SetFieldValue("ALMProject","$ToProjId");
    
	        if ($primary || $secondary || $roles || $primaryRole) {
		 		print("򡰽ɫֵ\n");
		 		if ($primary) {
	       	 		@ptypes = split(/\n/, $primary);
	                print("ѸƵɫΪ@ptypes\n");
		 			foreach (@ptypes) {
						$failure=$ent->AddFieldValue("ChildrenTypes_primary","$_");
	       			}  
		 		}
	        	if ($secondary) {
		       	   	@stypes = split(/\n/, $secondary);
			 		print("ѸƵĸɫΪ@stypes\n");
			 		foreach (@stypes) {
						$failure=$ent->AddFieldValue("ChildrenTYpes_secondary","$_");
		 	 		} 
	         	}
		 	 	if ($roles) {
	       	   	 	@AWCroles = split(/\n/, $roles);
				 	foreach (@AWCroles) {
						s/$FromProjId/$ToProjId/;
						print("ѸƵĽɫΪ$_\n");
						$failure=$ent->AddFieldValue("ALMRoles","$_");
		 	 		} 
	         	}
	         	if ($primaryRole) {
      			 	$primaryRole =~ s/$FromProjId/$ToProjId/;
		 		 	print("ѸƵ PrimaryRole Ϊ$primaryRole\n");
		 		 	$failure=$ent->SetFieldValue("PrimaryRole","$primaryRole");
	        	}
 
	   		}

		}
	}
	print(" ValidateEntities\n");
	
	############################
	# catch exceptions with eval
	eval{$strarray1 = $session->ValidateEntities($entityobjs);};
	my $size = grep(/\S/, @$strarray1); 
	if  ($@ || $size) {
			$session->RevertEntities($entityobjs);
			$ent=undef;
			$entityobjs=undef;
			CQPerlExt::CQSession_Unbuild($session);
			die "**  ValidateEntities յ$@ \nص״̬Ϊ@$strarray1\n";	
	}
	else
	{
		print(" CommitEntities\n");
		eval {	$strarray2  = $session->CommitEntities($entityobjs);};
		my $size = grep(/\S/, @$strarray2);
		if  ($@ || $size)  
		{
			$session->RevertEntities($entityobjs);
			$ent=undef;
			$entityobjs=undef;
			CQPerlExt::CQSession_Unbuild($session);
			die "**  CommitEntities յ$@ \nص״̬Ϊ@$strarray2\n";
		}
		else{
			print("CopyEntity ѳɹ\n");
			$ent=undef;
			$entityobjs=undef;
		}
	}
	print("\n$recordType ѳɹش $FromProjId Ƶ $ToProjId\n");
    return $result;
}
